==> spellathon-5.34.tar.gz: spellathon-5.34/src/logic/dict.cc <==
/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
# include <sys/stat.h>
# include <pwd.h>
#endif /* WIN32 */

#include <iostream>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <vector>

#ifndef SPTH_CL
# include <QProgressDialog>
# include <QFileInfo>
# include <QDateTime>
#endif /* SPTH_CL */

#include "dict.h"
#include "util.h"
#include "solver.h"

namespace Spellathon
{
	void
	Dict::dictOpen(const char *path)
	{
		char	line[MAX_LNLEN], *r; 
		int		nelem, slen, i;
		FILE	*fd;
	
		fd = fopen(path, "r");
		if (fd == 0) 
			spth_error(1, "Can't open dictionary: %s", path);
	
		nent = 0;
		nelem = 1<<14;
	
		dp = (dict_ent *)memalloc(nelem * sizeof(dict_ent));
		memset(dp, 0, nelem*sizeof(dict_ent));
	
		/* 
		 * read the wordlist and find all (appropriate) 
		 * words and store their signatures in mem.
		 */
again:
		while ( (r = fgets(line, MAX_LNLEN, fd)) != NULL) {
			slen = strlen(line);

			if (line[slen-1] != '\n') {
				fclose(fd);
				destroy();
				errno = EINVAL;
				spth_error(0, "%s: Unrecognized file type.\n\nDictionary must "
						   "be an ascii text file with the words seperated by newlines.", path);
			}

			line[--slen] = 0;

			if (line[slen-1] == '\r')
				line[--slen] = 0;
	
			if (slen < MIN_LETRS-1 || slen > MAX_LETRS)
				goto again;
	
			for (i = 0; i < slen; i++) 
				if(!islower(line[i]))
					goto again;
	
			if (nent >= nelem) {
				dp = (dict_ent *)memresize(dp, nelem * sizeof(dict_ent) * 2);
				memset(dp + nelem, 0, nelem * sizeof(dict_ent));
				nelem *= 2;
			}
	
			strcpy(dp[nent].word,line);
			sign(line, slen);
			strcpy(dp[nent++].sign,line);
		}
	
		if (ferror(fd)) {
			fclose(fd);
			destroy();
			spth_error(1, "%s: Read error", path);
		}
	
		fclose(fd);
	
		/* sort the dictionary */
		qsort((void *)dp, nent, sizeof(dict_ent), dictSortcmp);
	
		wp = (char (*)[8])memalloc(nent * 8);
		memset(wp, 0, nent * 8);
		for (int i = 0; i < nent; i++) 
			strcpy(wp[i], dp[i].word);

		qsort((void *)wp, nent, 8, dictSortWords);

		/* find strlen offsets */
	
		memset(off, 0, sizeof(off)); 

		off[3].s = off[3].e = -1;
		off[4].s = off[4].e = -1;
		off[5].s = off[5].e = -1;
		off[6].s = off[6].e = -1;
		off[7].s = off[7].e = -1;
		int curlen = 0;

		for (int i = 0; i < nent; i++) {
			int len = strlen(dp[i].sign);
			if (len > curlen) {
				off[len].s = i;
				curlen = len;
			}
			off[curlen].e = i;
		}
	}

	Dict::Dict() 
		: dp(0), wp(0), precomp_p(0)
	{ 
		dictbin[0] = '\0';
	}

	void
	Dict::init(const char *path, bool precomp) 
	{
		destroy();
		precomp_ = precomp;
		if (precomp_) {
			dictBinPath();
			dictPcOpen(path); 
		} else {
			dictOpen(path); 
		}
	}
	
	void
	Dict::dictBinPath()
	{
#ifdef WIN32
		strcpy(dictbin, getenv("APPDATA"));
		strcat(dictbin, "/"DICTBIN);
#else
		struct passwd *pw = getpwuid(getuid());
		if (pw == 0)
			spth_error(1, "Can't get user info");

		strcpy(dictbin, pw->pw_dir);
		strcat(dictbin, "/.spellathon/");
		if (access(dictbin, R_OK | W_OK) < 0) {
			if (mkdir(dictbin, 
			          S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) < 0)
 			       spth_error(1, "Can't create %s", dictbin);
		}
			
		strcat(dictbin, DICTBIN);
#endif /* WIN32 */
	}

	void
	Dict::dictPcOpen(const char *path)
	{
		FILE *binfl = fopen(dictbin, "rb");	
		if (binfl == 0) {
			if (errno == ENOENT) {
				dictPcCreate(path);
				return;
			} else {
 			       spth_error(1, "Can't open precomputed dictionary: %s", dictbin);
			}
		}
			

		fread(&precomp_hdr, sizeof(precomp_hdr_), 1, binfl);
#ifdef SPTH_CL 
# ifndef WIN32
		struct stat binp, dictp;
		stat(dictbin, &binp);
		stat(path, &dictp);
# endif
#endif /* SPTH_CL */
		if (strcmp(precomp_hdr.path, path) ||
		    strcmp(precomp_hdr.magic, MAGIC) 
#ifdef SPTH_CL 
# ifndef WIN32
		    || binp.st_mtime < dictp.st_mtime
# endif
#else
		    || QFileInfo(dictbin).lastModified().toTime_t() <
		    QFileInfo(path).lastModified().toTime_t()
#endif /* SPTH_CL */
		) {
			fclose(binfl);
			dictPcCreate(path);
			return;
		}

		nent = precomp_hdr.d_nent;
		memcpy(off, precomp_hdr.off, 8 * sizeof(off_));

		precomp_p = (precomp_ent*)memalloc(precomp_hdr.pc_nent *
			     sizeof(precomp_ent));

		dp = (dict_ent*)memalloc(precomp_hdr.d_nent * sizeof(dict_ent));

		if (fread(dp, sizeof(dict_ent), precomp_hdr.d_nent, binfl) 
		    < (unsigned int)precomp_hdr.d_nent) {
			fclose(binfl);
			destroy();
			spth_error(1, "%s: Read error", dictbin);
		}

		if (fread(precomp_p, sizeof(precomp_ent), precomp_hdr.pc_nent, binfl)			  < (unsigned int)precomp_hdr.pc_nent) {
			fclose(binfl);
			destroy();
			spth_error(1, "%s: Read error", dictbin);
		}

		wp = (char (*)[8])memalloc(nent * 8);
		memset(wp, 0, nent * 8);
		for (int i = 0; i < nent; i++) 
			strcpy(wp[i], dp[i].word);

		qsort((void *)wp, nent, 8, dictSortWords);

		fclose(binfl);
	}

	void
	Dict::dictPcCreate(const char *path)
	{
		int cnt = 0, n;

		dictOpen(path);
		Solver	sol(this);

		precomp_p = (precomp_ent*)memalloc(nent7() * 7 * sizeof(precomp_ent));

#ifndef SPTH_CL
		emit progressbarOpen("Precomputing Dictionary. Please wait...", "Dictionary", off[7].s , off[7].e);
#else
		std::cout << "Precomputing dictionary...  ";
#endif /* SPTH_CL */

		for (int i = off[7].s; i <= off[7].e; i++) {
			char sig[MAX_LETRS+1];
			strcpy(sig, dp[i].sign);

			int k = MAX_LETRS - 2;
			do {
				if ((n = sol.solve(sig)) > 0) {
					strcpy(precomp_p[cnt].sign, sig);	
					precomp_p[cnt++].nsol = n;
				}

				if (k < 0)
					break;

				/* try a different central letter */	
				swapch(&sig[MAX_LETRS-1], &sig[k--]);
			} while (1);

			if (!(i & ((1<<8) - 1))) 
#ifndef SPTH_CL
				emit progressbarUpdate(i);
#else
				spinner();
#endif /* SPTH_CL */
		}
#ifdef SPTH_CL
		printf("\bdone.\n");
#else
		emit progressbarClose();
#endif /* SPTH_CL */

		/* sort by number of solutions */
		qsort((void *)precomp_p, cnt, sizeof(precomp_ent), dictPcSortcmp);

		/* remove dup signs */
		if (cnt) {
			precomp_ent *f = precomp_p;
			precomp_ent *l = precomp_p + cnt;
			precomp_ent *r = f;
			while (++f != l) {
				if (strcmp(r->sign, f->sign))
					memmove(++r, f, sizeof(precomp_ent));
			}
			++r;
			cnt = r - precomp_p;
		}

		/* fill header */
		memset(&precomp_hdr, 0, sizeof(precomp_hdr_));
		precomp_hdr.d_nent = nent;
		precomp_hdr.pc_nent = cnt;
		memcpy(precomp_hdr.off, off, 8 * sizeof(struct off_));
		strncpy(precomp_hdr.path, path, MAX_PATH_LEN);
		precomp_hdr.path[MAX_PATH_LEN-1] = 0;
		strcpy(precomp_hdr.magic, MAGIC);
		
		FILE *binfl = fopen(dictbin, "wb");
		if (binfl == 0) {
			destroy();
			spth_error(1, "Can't open precomputed dictionary: %s", dictbin);
		}

		if (fwrite(&precomp_hdr, sizeof(precomp_hdr_), 1, binfl) < 1 ||
			fwrite(dp, sizeof(dict_ent), precomp_hdr.d_nent, binfl) < (unsigned int)precomp_hdr.d_nent ||
			fwrite(precomp_p, sizeof(precomp_ent), precomp_hdr.pc_nent, binfl) < (unsigned int)precomp_hdr.pc_nent
		   ) {
			fclose(binfl);
			destroy();
			spth_error(1, "%s: Write error", dictbin);
		}

		fclose(binfl);
	}

	/* binary search the dict. */
	/* performance critical routine */
	inline int
	Dict::bsearch(const char * sign, int slen) const
	{
		int l, h, m, r;
	
		l = off[slen].s;
		h = off[slen].e;
		
		if (l != -1 && h != -1) {
			while (h >= l) {
				m = (l + h) / 2;

				/* r = strcmp(sign, dp[m].sign) */
				r = 0;
				if (!strcmp_em(sign, dp[m].sign)) {
					int i = 0;
					if (sign[0] != dp[m].sign[0]) { goto out; }
					if (sign[1] != dp[m].sign[1]) { i = 1; goto out; }
					if (sign[2] != dp[m].sign[2]) { i = 2; goto out; }
					if (sign[3] != dp[m].sign[3]) { i = 3; goto out; }
					if (sign[4] != dp[m].sign[4]) { i = 4; goto out; }
					if (sign[5] != dp[m].sign[5]) { i = 5; goto out; }
					if (sign[6] != dp[m].sign[6]) { i = 6; }
out:
					r = sign[i] - dp[m].sign[i];
				}

				if (r < 0)
						h = m - 1;
				else if (r > 0)
						l = m + 1;
				else
					return m;
			}	
		} 

		return -1;
	}

	/* performance critical routine */
	bool
	Dict::matchWord(const char *word) const
	{
		int l, h, m, r;
	
		int slen = strlen(word);
		l = off[slen].s;
		h = off[slen].e;
		
		if (l != -1 && h != -1) {
			while (h >= l) {
				m = (l + h) / 2;

				r = 0;
				if (!strcmp_em(word, wp[m])) {
					int i = 0;
					if (word[0] != wp[m][0]) { goto out; }
					if (word[1] != wp[m][1]) { i = 1; goto out; }
					if (word[2] != wp[m][2]) { i = 2; goto out; }
					if (word[3] != wp[m][3]) { i = 3; goto out; }
					if (word[4] != wp[m][4]) { i = 4; goto out; }
					if (word[5] != wp[m][5]) { i = 5; goto out; }
					if (word[6] != wp[m][6]) { i = 6; }
out:
					r = word[i] - wp[m][i];
				}

				if (r < 0)
						h = m - 1;
				else if (r > 0)
						l = m + 1;
				else
					return true;
			}	
		} 

		return false;
	}
	
	
	void	
	Dict::matchSign(const char * sign, vector<char*> &w) const
	{
		int pos, npos, slen;
	
		slen = strlen(sign);
	
		/* binary search dictionary for matching signature. */
		pos = bsearch(sign, slen);
	
		if (pos != -1) {
			/* 
			 * search for all remaining duplicate signatures.
			 * the duplicates can be in contiguous locations
			 * on both the left and right sides of the index 
			 * returned by bsearch(). 
			 */
			npos = pos - 1;
			/* scan left */
			while (npos >= off[slen].s && strcmp_em(dp[npos].sign, sign)) {
				/* copy the word pointers into w[] */
				w.push_back(dp[npos--].word);
			}
			/* scan right */
			do {
				w.push_back(dp[pos].word);
			} while (++pos < off[slen].e && strcmp_em(dp[pos].sign, sign));
		}
	}
	
	/* sort by strlen() and then lexicographically */
	int
	Dict::dictSortcmp(const void *p1, const void *p2)
	{
		const char *s1 = ((const dict_ent *)p1)->sign;
		const char *s2 = ((const dict_ent *)p2)->sign;
		int r;
		
		r = strlen(s1) - strlen(s2);
		if (r == 0) 
			return strcmp(s1, s2);
		else 
			return r;
	}
	
	int
	Dict::dictSortWords(const void *p1, const void *p2)
	{
		const char *s1 = (const char *)p1;
		const char *s2 = (const char *)p2;
		int r;
		
		r = strlen(s1) - strlen(s2);
		if (r == 0) 
			return strcmp(s1, s2);
		else 
			return r;
	}
	
	int
	Dict::dictPcSortcmp(const void *p1, const void *p2)
	{
		const precomp_ent *s1 = (const precomp_ent *)p1;
		const precomp_ent *s2 = (const precomp_ent *)p2;
		
		int r = s1->nsol - s2->nsol;
		if (r == 0) 
			return strcmp(s1->sign, s2->sign);
		else 
			return r;
	}

	void
	Dict::destroy()
	{
		free(dp);
		dp = 0;
		free(wp);
		wp = 0;
		free(precomp_p);	
		precomp_p = 0;
	}
}

==> spellathon-5.40.tar.gz: spellathon-5.40/src/logic/dict.cc <==
/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
# include <sys/stat.h>
# include <pwd.h>
#endif /* WIN32 */

#include <iostream>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <vector>

#ifndef SPTH_CL
# include <QProgressDialog>
# include <QFileInfo>
# include <QDateTime>
#endif /* SPTH_CL */

#include "dict.h"
#include "util.h"
#include "solver.h"

namespace Spellathon
{
	void
	Dict::dictOpen(const char *path)
	{
		char	line[MAX_LNLEN], *r; 
		int		nelem, slen, i;
		FILE	*fd;
	
		fd = fopen(path, "r");
		if (fd == 0) 
			spth_error(1, "Can't open dictionary: %s", path);
	
		nent = 0;
		nelem = 1<<14;
	
		dp = (dict_ent *)memalloc(nelem * sizeof(dict_ent));
		memset(dp, 0, nelem*sizeof(dict_ent));
	
		/* 
		 * read the wordlist and find all (appropriate) 
		 * words and store their signatures in mem.
		 */
again:
		while ( (r = fgets(line, MAX_LNLEN, fd)) != NULL) {
			slen = strlen(line);

			if (line[slen-1] != '\n') {
				fclose(fd);
				destroy();
				errno = EINVAL;
				spth_error(0, "%s: Unrecognized file type.\n\nDictionary must "
						   "be an ascii text file with the words seperated by newlines.", path);
			}

			line[--slen] = 0;

			if (line[slen-1] == '\r')
				line[--slen] = 0;
	
			if (slen < MIN_LETRS-1 || slen > MAX_LETRS)
				goto again;
	
			for (i = 0; i < slen; i++) 
				if(!islower(line[i]))
					goto again;
	
			if (nent >= nelem) {
				dp = (dict_ent *)memresize(dp, nelem * sizeof(dict_ent) * 2);
				memset(dp + nelem, 0, nelem * sizeof(dict_ent));
				nelem *= 2;
			}
	
			strcpy(dp[nent].word,line);
			sign(line, slen);
			strcpy(dp[nent++].sign,line);
		}
	
		if (ferror(fd)) {
			fclose(fd);
			destroy();
			spth_error(1, "%s: Read error", path);
		}
	
		fclose(fd);
	
		/* sort the dictionary */
		qsort((void *)dp, nent, sizeof(dict_ent), dictSortcmp);
	
		wp = (char (*)[8])memalloc(nent * 8);
		memset(wp, 0, nent * 8);
		for (int i = 0; i < nent; i++) 
			strcpy(wp[i], dp[i].word);

		qsort((void *)wp, nent, 8, dictSortWords);

		/* find strlen offsets */
	
		memset(off, 0, sizeof(off)); 

		off[3].s = off[3].e = -1;
		off[4].s = off[4].e = -1;
		off[5].s = off[5].e = -1;
		off[6].s = off[6].e = -1;
		off[7].s = off[7].e = -1;
		int curlen = 0;

		for (int i = 0; i < nent; i++) {
			int len = strlen(dp[i].sign);
			if (len > curlen) {
				off[len].s = i;
				curlen = len;
			}
			off[curlen].e = i;
		}
	}

	Dict::Dict() 
		: dp(0), wp(0), precomp_p(0)
	{ 
		dictbin[0] = '\0';
	}

	void
	Dict::init(const char *path, bool precomp, bool plural_sieve) 
	{
		destroy();
		precomp_ = precomp;
		plural_sieve_ = plural_sieve;
		if (precomp_) {
			dictBinPath();
			dictPcOpen(path); 
		} else {
			dictOpen(path); 
		}
	}
	
	void
	Dict::dictBinPath()
	{
#ifdef WIN32
		strcpy(dictbin, getenv("APPDATA"));
		strcat(dictbin, "/"DICTBIN);
#else
		struct passwd *pw = getpwuid(getuid());
		if (pw == 0)
			spth_error(1, "Can't get user info");

		strcpy(dictbin, pw->pw_dir);
		strcat(dictbin, "/.spellathon/");
		if (access(dictbin, R_OK | W_OK) < 0) {
			if (mkdir(dictbin, 
			          S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) < 0)
 			       spth_error(1, "Can't create %s", dictbin);
		}
			
		strcat(dictbin, DICTBIN);
#endif /* WIN32 */
	}

	void
	Dict::dictPcOpen(const char *path)
	{
		FILE *binfl = fopen(dictbin, "rb");	
		if (binfl == 0) {
			if (errno == ENOENT) {
				dictPcCreate(path);
				return;
			} else {
 			       spth_error(1, "Can't open precomputed dictionary: %s", dictbin);
			}
		}
			

		fread(&precomp_hdr, sizeof(precomp_hdr_), 1, binfl);
#ifdef SPTH_CL 
# ifndef WIN32
		struct stat binp, dictp;
		stat(dictbin, &binp);
		stat(path, &dictp);
# endif
#endif /* SPTH_CL */
		if (strcmp(precomp_hdr.path, path) ||
		    strcmp(precomp_hdr.magic, MAGIC) ||
			(precomp_hdr.plural_sieve > 0) != plural_sieve_
#ifdef SPTH_CL 
# ifndef WIN32
		    || binp.st_mtime < dictp.st_mtime
# endif
#else
		    || QFileInfo(dictbin).lastModified().toTime_t() <
		    QFileInfo(path).lastModified().toTime_t()
#endif /* SPTH_CL */
		) {
			fclose(binfl);
			dictPcCreate(path);
			return;
		}

		nent = precomp_hdr.d_nent;
		memcpy(off, precomp_hdr.off, 8 * sizeof(off_));

		precomp_p = (precomp_ent*)memalloc(precomp_hdr.pc_nent *
			     sizeof(precomp_ent));

		dp = (dict_ent*)memalloc(precomp_hdr.d_nent * sizeof(dict_ent));

		if (fread(dp, sizeof(dict_ent), precomp_hdr.d_nent, binfl) 
		    < (unsigned int)precomp_hdr.d_nent) {
			fclose(binfl);
			destroy();
			spth_error(1, "%s: Read error", dictbin);
		}

		if (fread(precomp_p, sizeof(precomp_ent), precomp_hdr.pc_nent, binfl)			  < (unsigned int)precomp_hdr.pc_nent) {
			fclose(binfl);
			destroy();
			spth_error(1, "%s: Read error", dictbin);
		}

		wp = (char (*)[8])memalloc(nent * 8);
		memset(wp, 0, nent * 8);
		for (int i = 0; i < nent; i++) 
			strcpy(wp[i], dp[i].word);

		qsort((void *)wp, nent, 8, dictSortWords);

		fclose(binfl);
	}

	void
	Dict::dictPcCreate(const char *path)
	{
		int cnt = 0, n;

		dictOpen(path);
		Solver	sol(this, plural_sieve_);

		precomp_p = (precomp_ent*)memalloc(nent7() * 7 * sizeof(precomp_ent));

#ifndef SPTH_CL
		emit progressbarOpen("Precomputing Dictionary. Please wait...", "Dictionary", off[7].s , off[7].e);
#else
		std::cout << "Precomputing dictionary...  ";
#endif /* SPTH_CL */

		for (int i = off[7].s; i <= off[7].e; i++) {
			char sig[MAX_LETRS+1];
			strcpy(sig, dp[i].sign);

			int k = MAX_LETRS - 2;
			do {
				if ((n = sol.solve(sig)) > 0) {
					strcpy(precomp_p[cnt].sign, sig);	
					precomp_p[cnt++].nsol = n;
				}

				if (k < 0)
					break;

				/* try a different central letter */	
				swapch(&sig[MAX_LETRS-1], &sig[k--]);
			} while (1);

			if (!(i & ((1<<8) - 1))) 
#ifndef SPTH_CL
				emit progressbarUpdate(i);
#else
				spinner();
#endif /* SPTH_CL */
		}
#ifdef SPTH_CL
		printf("\bdone.\n");
#else
		emit progressbarClose();
#endif /* SPTH_CL */

		/* sort by number of solutions */
		qsort((void *)precomp_p, cnt, sizeof(precomp_ent), dictPcSortcmp);

		/* remove dup signs */
		if (cnt) {
			precomp_ent *f = precomp_p;
			precomp_ent *l = precomp_p + cnt;
			precomp_ent *r = f;
			while (++f != l) {
				if (strcmp(r->sign, f->sign))
					memmove(++r, f, sizeof(precomp_ent));
			}
			++r;
			cnt = r - precomp_p;
		}

		/* fill header */
		memset(&precomp_hdr, 0, sizeof(precomp_hdr_));
		precomp_hdr.d_nent = nent;
		precomp_hdr.pc_nent = cnt;
		memcpy(precomp_hdr.off, off, 8 * sizeof(struct off_));
		strncpy(precomp_hdr.path, path, MAX_PATH_LEN);
		precomp_hdr.path[MAX_PATH_LEN-1] = 0;
		strcpy(precomp_hdr.magic, MAGIC);
		precomp_hdr.plural_sieve = plural_sieve_;
		
		FILE *binfl = fopen(dictbin, "wb");
		if (binfl == 0) {
			destroy();
			spth_error(1, "Can't open precomputed dictionary: %s", dictbin);
		}

		if (fwrite(&precomp_hdr, sizeof(precomp_hdr_), 1, binfl) < 1 ||
			fwrite(dp, sizeof(dict_ent), precomp_hdr.d_nent, binfl) < (unsigned int)precomp_hdr.d_nent ||
			fwrite(precomp_p, sizeof(precomp_ent), precomp_hdr.pc_nent, binfl) < (unsigned int)precomp_hdr.pc_nent
		   ) {
			fclose(binfl);
			destroy();
			spth_error(1, "%s: Write error", dictbin);
		}

		fclose(binfl);
	}

	/* binary search the dict. */
	/* performance critical routine */
	inline int
	Dict::bsearch(const char * sign, int slen) const
	{
		int l, h, m, r;
	
		l = off[slen].s;
		h = off[slen].e;
		
		if (l != -1 && h != -1) {
			while (h >= l) {
				m = (l + h) / 2;

				/* r = strcmp(sign, dp[m].sign) */
				r = 0;
				if (!strcmp_em(sign, dp[m].sign)) {
					int i = 0;
					if (sign[0] != dp[m].sign[0]) { goto out; }
					if (sign[1] != dp[m].sign[1]) { i = 1; goto out; }
					if (sign[2] != dp[m].sign[2]) { i = 2; goto out; }
					if (sign[3] != dp[m].sign[3]) { i = 3; goto out; }
					if (sign[4] != dp[m].sign[4]) { i = 4; goto out; }
					if (sign[5] != dp[m].sign[5]) { i = 5; goto out; }
					if (sign[6] != dp[m].sign[6]) { i = 6; }
out:
					r = sign[i] - dp[m].sign[i];
				}

				if (r < 0)
						h = m - 1;
				else if (r > 0)
						l = m + 1;
				else
					return m;
			}	
		} 

		return -1;
	}

	/* performance critical routine */
	bool
	Dict::matchWord(const char *word) const
	{
		int l, h, m, r;
	
		int slen = strlen(word);
		l = off[slen].s;
		h = off[slen].e;
		
		if (l != -1 && h != -1) {
			while (h >= l) {
				m = (l + h) / 2;

				r = 0;
				if (!strcmp_em(word, wp[m])) {
					int i = 0;
					if (word[0] != wp[m][0]) { goto out; }
					if (word[1] != wp[m][1]) { i = 1; goto out; }
					if (word[2] != wp[m][2]) { i = 2; goto out; }
					if (word[3] != wp[m][3]) { i = 3; goto out; }
					if (word[4] != wp[m][4]) { i = 4; goto out; }
					if (word[5] != wp[m][5]) { i = 5; goto out; }
					if (word[6] != wp[m][6]) { i = 6; }
out:
					r = word[i] - wp[m][i];
				}

				if (r < 0)
						h = m - 1;
				else if (r > 0)
						l = m + 1;
				else
					return true;
			}	
		} 

		return false;
	}
	
	
	void	
	Dict::matchSign(const char * sign, vector<char*> &w) const
	{
		int pos, npos, slen;
	
		slen = strlen(sign);
	
		/* binary search dictionary for matching signature. */
		pos = bsearch(sign, slen);
	
		if (pos != -1) {
			/* 
			 * search for all remaining duplicate signatures.
			 * the duplicates can be in contiguous locations
			 * on both the left and right sides of the index 
			 * returned by bsearch(). 
			 */
			npos = pos - 1;
			/* scan left */
			while (npos >= off[slen].s && strcmp_em(dp[npos].sign, sign)) {
				/* copy the word pointers into w[] */
				w.push_back(dp[npos--].word);
			}
			/* scan right */
			do {
				w.push_back(dp[pos].word);
			} while (++pos < off[slen].e && strcmp_em(dp[pos].sign, sign));
		}
	}
	
	/* sort by strlen() and then lexicographically */
	int
	Dict::dictSortcmp(const void *p1, const void *p2)
	{
		const char *s1 = ((const dict_ent *)p1)->sign;
		const char *s2 = ((const dict_ent *)p2)->sign;
		int r;
		
		r = strlen(s1) - strlen(s2);
		if (r == 0) 
			return strcmp(s1, s2);
		else 
			return r;
	}
	
	int
	Dict::dictSortWords(const void *p1, const void *p2)
	{
		const char *s1 = (const char *)p1;
		const char *s2 = (const char *)p2;
		int r;
		
		r = strlen(s1) - strlen(s2);
		if (r == 0) 
			return strcmp(s1, s2);
		else 
			return r;
	}
	
	int
	Dict::dictPcSortcmp(const void *p1, const void *p2)
	{
		const precomp_ent *s1 = (const precomp_ent *)p1;
		const precomp_ent *s2 = (const precomp_ent *)p2;
		
		int r = s1->nsol - s2->nsol;
		if (r == 0) 
			return strcmp(s1->sign, s2->sign);
		else 
			return r;
	}

	void
	Dict::destroy()
	{
		free(dp);
		dp = 0;
		free(wp);
		wp = 0;
		free(precomp_p);	
		precomp_p = 0;
	}
}
